# MiloDB by FrozenWolf is marked with CC0 1.0.
# To view a copy of this license, visit http://creativecommons.org/publicdomain/zero/1.0
import unittest
from unittest.mock import Mock, PropertyMock
from milodb_client.commands.sort_keys import AUTHOR_ID_SORT_KEY, DATE_SORT_KEY, MATCH_COUNT_SORT_KEY, OrderedSortKey, RATING_SORT_KEY, TEASE_ID_SORT_KEY, TITLE_SORT_KEY, TOTM_SORT_KEY, TYPE_SORT_KEY
from milodb_client.commands.tease_match_sorter import TeaseMatchSorter
from milodb_client.query.field_match import IFieldMatch
from milodb_client.query.tease_match import TeaseMatch
from milodb_client_test.test import fake_teases
from milodb_client_test.test.matchers_teases import is_equal_list_of_tease_matches

_DEFAULT_LIST_OF_TEASE_MATCHES: list[TeaseMatch] = [
    TeaseMatch(1, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
    TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
    TeaseMatch(3, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
    TeaseMatch(4, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
    TeaseMatch(5, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [])
]

class TestTeaseMatchSorter(unittest.TestCase):
    def test_sort_empty_list_with_no_keys_returns_empty_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort([], [])
        self.assertListEqual([], sorted_list)

    def test_sort_empty_list_with_some_keys_returns_empty_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort([], [OrderedSortKey(TEASE_ID_SORT_KEY, True), OrderedSortKey(DATE_SORT_KEY, False)])
        self.assertListEqual([], sorted_list)

    def test_sort_empty_list_with_no_keys_and_some_defaults_returns_empty_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([OrderedSortKey(TEASE_ID_SORT_KEY, True), OrderedSortKey(DATE_SORT_KEY, False)]).sort([], [])
        self.assertListEqual([], sorted_list)

    def test_sort_with_no_keys_returns_same_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, _DEFAULT_LIST_OF_TEASE_MATCHES))

    def test_sort_by_tease_id_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TEASE_ID_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_sort_by_tease_id_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TEASE_ID_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(2, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, [])
        ]))

    def test_sort_by_author_id_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(AUTHOR_ID_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_sort_by_author_id_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(AUTHOR_ID_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, [])
        ]))

    def test_sort_by_date_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(DATE_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_sort_by_date_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(DATE_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(2, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, [])
        ]))

    def test_sort_by_title_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TITLE_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(4, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(5, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [])
        ]))

    def test_sort_by_title_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TITLE_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(3, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, [])
        ]))

    def test_sort_by_rating_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(RATING_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(2, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [])
        ]))

    def test_sort_by_rating_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(RATING_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_sort_by_type_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TYPE_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(3, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, [])
        ]))

    def test_sort_by_type_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TYPE_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(5, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [])
        ]))

    def test_sort_by_totm_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TOTM_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(4, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, [])
        ]))

    def test_sort_by_totm_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(TOTM_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ], True))

    def test_sort_by_match_count_ascending_returns_sorted_list(self) -> None:
        mock_indices: Mock = Mock(IFieldMatch.Indices)
        list_of_one_field_match_with_four_indices: list[Mock] = [Mock(IFieldMatch)]
        type(list_of_one_field_match_with_four_indices[0]).list_of_indices = PropertyMock(return_value=[mock_indices, mock_indices, mock_indices, mock_indices])
        list_of_two_field_matches_with_five_indices: list[Mock] = [Mock(IFieldMatch), Mock(IFieldMatch)]
        type(list_of_two_field_matches_with_five_indices[0]).list_of_indices = PropertyMock(return_value=[mock_indices, mock_indices])
        type(list_of_two_field_matches_with_five_indices[1]).list_of_indices = PropertyMock(return_value=[mock_indices, mock_indices, mock_indices])
        list_of_three_field_matches_with_three_indices: list[Mock] = [Mock(IFieldMatch), Mock(IFieldMatch), Mock(IFieldMatch)]
        type(list_of_three_field_matches_with_three_indices[0]).list_of_indices = PropertyMock(return_value=[mock_indices])
        type(list_of_three_field_matches_with_three_indices[1]).list_of_indices = PropertyMock(return_value=[mock_indices])
        type(list_of_three_field_matches_with_three_indices[2]).list_of_indices = PropertyMock(return_value=[mock_indices])
        list_of_tease_matches: list[TeaseMatch] = [
            TeaseMatch(1, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, [*list_of_two_field_matches_with_five_indices]),
            TeaseMatch(3, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(4, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [*list_of_three_field_matches_with_three_indices]),
            TeaseMatch(5, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [*list_of_one_field_match_with_four_indices])
        ]
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(list_of_tease_matches, [OrderedSortKey(MATCH_COUNT_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [*list_of_three_field_matches_with_three_indices]),
            TeaseMatch(4, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [*list_of_one_field_match_with_four_indices]),
            TeaseMatch(5, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, [*list_of_two_field_matches_with_five_indices])
        ]))

    def test_sort_by_match_count_descending_returns_sorted_list(self) -> None:
        mock_indices: Mock = Mock(IFieldMatch.Indices)
        list_of_one_field_match_with_four_indices: list[Mock] = [Mock(IFieldMatch)]
        type(list_of_one_field_match_with_four_indices[0]).list_of_indices = PropertyMock(return_value=[mock_indices, mock_indices, mock_indices, mock_indices])
        list_of_two_field_matches_with_five_indices: list[Mock] = [Mock(IFieldMatch), Mock(IFieldMatch)]
        type(list_of_two_field_matches_with_five_indices[0]).list_of_indices = PropertyMock(return_value=[mock_indices, mock_indices])
        type(list_of_two_field_matches_with_five_indices[1]).list_of_indices = PropertyMock(return_value=[mock_indices, mock_indices, mock_indices])
        list_of_three_field_matches_with_three_indices: list[Mock] = [Mock(IFieldMatch), Mock(IFieldMatch), Mock(IFieldMatch)]
        type(list_of_three_field_matches_with_three_indices[0]).list_of_indices = PropertyMock(return_value=[mock_indices])
        type(list_of_three_field_matches_with_three_indices[1]).list_of_indices = PropertyMock(return_value=[mock_indices])
        type(list_of_three_field_matches_with_three_indices[2]).list_of_indices = PropertyMock(return_value=[mock_indices])
        list_of_tease_matches: list[TeaseMatch] = [
            TeaseMatch(1, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, [*list_of_two_field_matches_with_five_indices]),
            TeaseMatch(3, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(4, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [*list_of_three_field_matches_with_three_indices]),
            TeaseMatch(5, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [*list_of_one_field_match_with_four_indices])
        ]
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(list_of_tease_matches, [OrderedSortKey(MATCH_COUNT_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, [*list_of_two_field_matches_with_five_indices]),
            TeaseMatch(2, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, [*list_of_one_field_match_with_four_indices]),
            TeaseMatch(3, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [*list_of_three_field_matches_with_three_indices]),
            TeaseMatch(4, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, [])
        ]))

    def test_sort_by_author_id_ascending_and_tease_id_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(AUTHOR_ID_SORT_KEY, True), OrderedSortKey(TEASE_ID_SORT_KEY, False)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_sort_by_author_id_ascending_and_tease_id_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(AUTHOR_ID_SORT_KEY, True), OrderedSortKey(TEASE_ID_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_default_sort_by_author_id_ascending_and_tease_id_descending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([OrderedSortKey(AUTHOR_ID_SORT_KEY, True), OrderedSortKey(TEASE_ID_SORT_KEY, False)]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))

    def test_default_tease_id_descending_and_specified_sort_by_author_id_ascending_returns_sorted_list(self) -> None:
        sorted_list: list[TeaseMatch] = TeaseMatchSorter([OrderedSortKey(TEASE_ID_SORT_KEY, False)]).sort(_DEFAULT_LIST_OF_TEASE_MATCHES, [OrderedSortKey(AUTHOR_ID_SORT_KEY, True)])
        self.assertTrue(is_equal_list_of_tease_matches(sorted_list, [
            TeaseMatch(1, fake_teases.TEASE_ID_5678_AUTHOR_ID_4567, []),
            TeaseMatch(2, fake_teases.TEASE_ID_1000_AUTHOR_ID_4567, []),
            TeaseMatch(3, fake_teases.TEASE_ID_2000_AUTHOR_ID_6431, []),
            TeaseMatch(4, fake_teases.TEASE_ID_1234_AUTHOR_ID_6431, []),
            TeaseMatch(5, fake_teases.TEASE_ID_8721_AUTHOR_ID_7213, [])
        ]))
