# MiloDB by FrozenWolf is marked with CC0 1.0.
# To view a copy of this license, visit http://creativecommons.org/publicdomain/zero/1.0
import unittest
from pathlib import Path
from unittest.mock import Mock, PropertyMock, patch
from milodb_client.updater.i_file_hasher import HashError, HashFileNotFoundError, IFileHasher
from milodb_client.updater.i_file_tester import FileStat, IFileTester
from milodb_client.updater.manifest.common_manifest import IConfigFile, IHexDigest, ILaunchFile, IVersionNumber
from milodb_client.updater.manifest.local_manifest import ILocalManifest
from milodb_client.updater.manifest.update_directory import IAsset, IAssetRegister
from milodb_client.updater.manifest.version_manifest import ICoreFile, IVariant, IVersion, IVersionManifest
from milodb_client.updater.manifest.version_number import VersionNumber
from milodb_client.updater.update_strategy import BackupFileSet, ConfigFileChanges, CoreFileChanges, FileRename, UpdateStrategy, UpdateStrategyError, determine_config_file_changes, determine_core_file_changes, determine_files_to_backup, determine_newer_versions, determine_update_strategy, find_target_variant_and_version
from milodb_client_test.test.fake_data import FAKE_HEXDIGEST_A, FAKE_HEXDIGEST_B, FAKE_HEXDIGEST_C, FAKE_HEXDIGEST_D, FAKE_HEXDIGEST_E, FAKE_HEXDIGEST_F, FAKE_HEXDIGEST_G, FAKE_HEXDIGEST_H, FAKE_HEXDIGEST_I
from milodb_common_test.test import mock_path
from milodb_common_test.test.strict_mock import StrictMock

class TestFindTargetVariantAndVersion(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.variant_windows = StrictMock(IVariant)
        self.variant_linux = StrictMock(IVariant)
        self.variant_vms = StrictMock(IVariant)
        self.variant_windows.name = PropertyMock(return_value='windows')
        self.variant_linux.name = PropertyMock(return_value='linux')
        self.variant_vms.name = PropertyMock(return_value='vms')

        self.version_number_1 = VersionNumber.parse('1.0.0')
        self.version_number_2 = VersionNumber.parse('2.0.0')
        self.version_number_3 = VersionNumber.parse('3.0.0')
        self.unknown_version_number = VersionNumber.parse('2.1.3')

        self.version_1 = StrictMock(IVersion)
        self.version_2 = StrictMock(IVersion)
        self.version_3 = StrictMock(IVersion)
        self.version_1.number = PropertyMock(return_value=self.version_number_1)
        self.version_2.number = PropertyMock(return_value=self.version_number_2)
        self.version_3.number = PropertyMock(return_value=self.version_number_3)

        self.variant_windows.versions = PropertyMock(return_value={})
        self.variant_linux.versions = PropertyMock(return_value={
            self.version_3.number: self.version_3,
            self.version_1.number: self.version_1,
            self.version_2.number: self.version_2
        })
        self.variant_vms.versions = PropertyMock(return_value={})

        self.map_of_variants: dict[str, Mock] = {
            self.variant_windows.name: self.variant_windows,
            self.variant_linux.name: self.variant_linux,
            self.variant_vms.name: self.variant_vms
        }

    def test_finds_latest_version_when_version_number_not_specified(self) -> None:
        requested_variant_name: str = 'linux'
        requested_version_number: IVersionNumber | None = None

        target_variant: IVariant
        target_version: IVersion
        target_variant, target_version = find_target_variant_and_version(requested_version_number, requested_variant_name, self.map_of_variants)

        self.assertIs(self.variant_linux, target_variant)
        self.assertIs(self.version_3, target_version)

    def test_finds_specified_version_when_specified(self) -> None:
        requested_variant_name: str = 'linux'
        requested_version_number: IVersionNumber | None = self.version_number_1

        target_variant: IVariant
        target_version: IVersion
        target_variant, target_version = find_target_variant_and_version(requested_version_number, requested_variant_name, self.map_of_variants)

        self.assertIs(self.variant_linux, target_variant)
        self.assertIs(self.version_1, target_version)

    def test_raises_error_when_specified_variant_does_not_exist(self) -> None:
        requested_variant_name: str = 'darwin'
        requested_version_number: IVersionNumber | None = None

        with self.assertRaises(UpdateStrategyError) as ex:
            find_target_variant_and_version(requested_version_number, requested_variant_name, self.map_of_variants)

        self.assertEqual(f"Requested variant '{requested_variant_name}' does not exist in the version manifest", str(ex.exception))

    def test_raises_error_when_specified_version_does_not_exist_within_variant(self) -> None:
        requested_variant_name: str = 'linux'
        requested_version_number: IVersionNumber | None = self.unknown_version_number

        with self.assertRaises(UpdateStrategyError) as ex:
            find_target_variant_and_version(requested_version_number, requested_variant_name, self.map_of_variants)

        self.assertEqual(f"Requested version '{requested_version_number}' in variant '{requested_variant_name}' does not exist in the version manifest", str(ex.exception))

    def test_raises_error_when_specified_variant_has_no_versions(self) -> None:
        requested_variant_name: str = 'windows'
        requested_version_number: IVersionNumber | None = None

        with self.assertRaises(UpdateStrategyError) as ex:
            find_target_variant_and_version(requested_version_number, requested_variant_name, self.map_of_variants)

        self.assertEqual(f"Requested variant '{requested_variant_name}' has no versions in the version manifest", str(ex.exception))

class TestDetermineCoreFileChanges(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.file_hasher = StrictMock(IFileHasher)

    def test_no_local_files_and_no_target_files_returns_no_changes(self) -> None:
        list_of_local_files: list[str] = []
        map_of_target_core_files: dict[str, ICoreFile] = {}
        set_of_asset_digests: set[IHexDigest] = set()

        core_file_changes: CoreFileChanges = determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.assertFalse(core_file_changes.list_of_new_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_missing_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_changed_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_unchanged_files)
        self.assertFalse(core_file_changes.list_of_files_that_are_deprecated)

    def test_two_local_files_and_no_target_files_returns_files_to_remove(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        map_of_target_core_files: dict[str, ICoreFile] = {}
        set_of_asset_digests: set[IHexDigest] = set()

        core_file_changes: CoreFileChanges = determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.assertFalse(core_file_changes.list_of_new_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_missing_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_changed_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_unchanged_files)
        self.assertListEqual(list_of_local_files, core_file_changes.list_of_files_that_are_deprecated)

    def test_two_local_files_and_one_target_file_with_same_name_and_digest_returns_one_unchanged_file(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        core_file = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='bar.dat'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_A)
        )
        map_of_target_core_files: dict[str, ICoreFile] = {
            core_file.filename: core_file
        }
        set_of_asset_digests: set[IHexDigest] = { FAKE_HEXDIGEST_A }
        self.file_hasher.hash_file = Mock(return_value=FAKE_HEXDIGEST_A.digest_bytes)

        core_file_changes: CoreFileChanges = determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.file_hasher.hash_file.assert_called_once_with('bar.dat')
        self.assertFalse(core_file_changes.list_of_new_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_missing_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_changed_files_to_acquire)
        self.assertListEqual([core_file], core_file_changes.list_of_unchanged_files)
        self.assertListEqual(['foo.txt'], core_file_changes.list_of_files_that_are_deprecated)

    def test_two_local_files_and_one_target_file_with_same_name_and_different_digest_returns_changed_file_to_acquire(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        core_file = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='bar.dat'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_A)
        )
        map_of_target_core_files: dict[str, ICoreFile] = {
            core_file.filename: core_file
        }
        set_of_asset_digests: set[IHexDigest] = { FAKE_HEXDIGEST_A }
        self.file_hasher.hash_file = Mock(return_value=FAKE_HEXDIGEST_B.digest_bytes)

        core_file_changes: CoreFileChanges = determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.file_hasher.hash_file.assert_called_once_with('bar.dat')
        self.assertFalse(core_file_changes.list_of_new_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_missing_files_to_acquire)
        self.assertListEqual([core_file], core_file_changes.list_of_changed_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_unchanged_files)
        self.assertListEqual(['foo.txt'], core_file_changes.list_of_files_that_are_deprecated)

    def test_two_local_files_and_one_target_file_with_different_name_and_different_digest_returns_new_file_to_acquire(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        core_file = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='fruit.id'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_A)
        )
        map_of_target_core_files: dict[str, ICoreFile] = {
            core_file.filename: core_file
        }
        set_of_asset_digests: set[IHexDigest] = { FAKE_HEXDIGEST_A }

        core_file_changes: CoreFileChanges = determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.assertListEqual([core_file], core_file_changes.list_of_new_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_missing_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_changed_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_unchanged_files)
        self.assertCountEqual(['foo.txt', 'bar.dat'], core_file_changes.list_of_files_that_are_deprecated)

    def test_two_local_files_where_one_is_missing_and_matching_target_files_with_same_digests_returns_missing_file_to_acquire(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        core_file_1 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='foo.txt'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_A)
        )
        core_file_2 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='bar.dat'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_B)
        )
        map_of_target_core_files: dict[str, ICoreFile] = {
            core_file_1.filename: core_file_1,
            core_file_2.filename: core_file_2
        }
        set_of_asset_digests: set[IHexDigest] = { FAKE_HEXDIGEST_A, FAKE_HEXDIGEST_B }
        def on_hash_file(filename: str) -> bytes:
            if filename == 'foo.txt':
                raise HashFileNotFoundError()
            if filename == 'bar.dat':
                return FAKE_HEXDIGEST_B.digest_bytes
            self.fail(f"Unexpected call to hash_file('{filename}')")
            return bytes()
        self.file_hasher.hash_file = Mock(side_effect=on_hash_file)

        core_file_changes: CoreFileChanges = determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.assertFalse(core_file_changes.list_of_new_files_to_acquire)
        self.assertListEqual([core_file_1], core_file_changes.list_of_missing_files_to_acquire)
        self.assertFalse(core_file_changes.list_of_changed_files_to_acquire)
        self.assertListEqual([core_file_2], core_file_changes.list_of_unchanged_files)
        self.assertFalse(core_file_changes.list_of_files_that_are_deprecated)

    def test_error_raised_when_a_hash_file_error_occurs(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        core_file = StrictMock(ICoreFile)
        core_file.filename = PropertyMock(return_value='bar.dat')
        core_file.digest = PropertyMock(return_value=FAKE_HEXDIGEST_B)
        map_of_target_core_files: dict[str, ICoreFile] = {
            core_file.filename: core_file
        }
        set_of_asset_digests: set[IHexDigest] = { FAKE_HEXDIGEST_B }
        self.file_hasher.hash_file = Mock(side_effect=HashError('test message'))

        with self.assertRaises(UpdateStrategyError) as ex:
            determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.assertEqual('test message', str(ex.exception))

    def test_error_raised_when_a_digest_is_not_in_the_asset_register(self) -> None:
        list_of_local_files: list[str] = ['foo.txt', 'bar.dat']
        core_file = StrictMock(ICoreFile)
        core_file.filename = PropertyMock(return_value='bar.dat')
        core_file.digest = PropertyMock(return_value=FAKE_HEXDIGEST_B)
        map_of_target_core_files: dict[str, ICoreFile] = {
            core_file.filename: core_file
        }
        set_of_asset_digests: set[IHexDigest] = { FAKE_HEXDIGEST_A }
        self.file_hasher.hash_file = Mock(side_effect=HashError('test message'))

        with self.assertRaises(UpdateStrategyError) as ex:
            determine_core_file_changes(list_of_local_files, map_of_target_core_files, set_of_asset_digests, self.file_hasher)

        self.assertEqual(f"Core file 'bar.dat' with digest '{FAKE_HEXDIGEST_B}' does not appear in the asset register", str(ex.exception))

class TestDetermineConfigFileChanges(unittest.TestCase):
    def test_no_local_files_and_no_target_files(self) -> None:
        """
        ### Local
        * (none)
        ### Target
        * (none)
        ### Expected
        * (no changes)
        """
        map_of_local_config_files: dict[int, IConfigFile] = {}
        map_of_target_config_files: dict[int, IConfigFile] = {}

        config_file_changes: ConfigFileChanges = determine_config_file_changes(map_of_local_config_files, map_of_target_config_files)

        self.assertFalse(config_file_changes.list_of_files_that_are_deprecated)
        self.assertFalse(config_file_changes.list_of_files_to_rename)
        self.assertFalse(config_file_changes.list_of_files_to_leave)
        self.assertFalse(config_file_changes.list_of_new_files)

    def test_two_local_files_and_no_target_files(self) -> None:
        """
        ### Local
        * 23: test.dat
        * 8: another.txt
        ### Target
        * (none)
        ### Expected
        * files_to_remove = test.dat, another.txt
        """
        local_file_1 = StrictMock(IConfigFile)
        local_file_1.config_key = PropertyMock(return_value=23)
        local_file_1.filename = PropertyMock(return_value='test.dat')
        local_file_2 = StrictMock(IConfigFile)
        local_file_2.config_key = PropertyMock(return_value=8)
        local_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_local_config_files: dict[int, IConfigFile] = {
            local_file_1.config_key: local_file_1,
            local_file_2.config_key: local_file_2
        }
        map_of_target_config_files: dict[int, IConfigFile] = {}

        config_file_changes: ConfigFileChanges = determine_config_file_changes(map_of_local_config_files, map_of_target_config_files)

        self.assertCountEqual(['test.dat', 'another.txt'], config_file_changes.list_of_files_that_are_deprecated)
        self.assertFalse(config_file_changes.list_of_files_to_rename)
        self.assertFalse(config_file_changes.list_of_files_to_leave)
        self.assertFalse(config_file_changes.list_of_new_files)

    def test_no_local_files_and_two_target_files(self) -> None:
        """
        ### Local
        * (none)
        ### Target
        * 23: test.dat
        * 8: another.txt
        ### Expected
        * new_files = test.dat, another.txt
        """
        map_of_local_config_files: dict[int, IConfigFile] = {}
        target_file_1 = StrictMock(IConfigFile)
        target_file_1.config_key = PropertyMock(return_value=23)
        target_file_1.filename = PropertyMock(return_value='test.dat')
        target_file_2 = StrictMock(IConfigFile)
        target_file_2.config_key = PropertyMock(return_value=8)
        target_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_target_config_files: dict[int, IConfigFile] = {
            target_file_1.config_key: target_file_1,
            target_file_2.config_key: target_file_2
        }

        config_file_changes: ConfigFileChanges = determine_config_file_changes(map_of_local_config_files, map_of_target_config_files)

        self.assertFalse(config_file_changes.list_of_files_that_are_deprecated)
        self.assertFalse(config_file_changes.list_of_files_to_rename)
        self.assertFalse(config_file_changes.list_of_files_to_leave)
        self.assertCountEqual(['test.dat', 'another.txt'], config_file_changes.list_of_new_files)

    def test_two_local_files_and_two_target_files_with_same_names_and_keys(self) -> None:
        """
        ### Local
        * 23: test.dat
        * 8: another.txt
        ### Target
        * 23: test.dat
        * 8: another.txt
        ### Expected
        * files_to_leave = test.dat, another.txt
        """
        local_file_1 = StrictMock(IConfigFile)
        local_file_1.config_key = PropertyMock(return_value=23)
        local_file_1.filename = PropertyMock(return_value='test.dat')
        local_file_2 = StrictMock(IConfigFile)
        local_file_2.config_key = PropertyMock(return_value=8)
        local_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_local_config_files: dict[int, IConfigFile] = {
            local_file_1.config_key: local_file_1,
            local_file_2.config_key: local_file_2
        }
        target_file_1 = StrictMock(IConfigFile)
        target_file_1.config_key = PropertyMock(return_value=23)
        target_file_1.filename = PropertyMock(return_value='test.dat')
        target_file_2 = StrictMock(IConfigFile)
        target_file_2.config_key = PropertyMock(return_value=8)
        target_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_target_config_files: dict[int, IConfigFile] = {
            target_file_1.config_key: target_file_1,
            target_file_2.config_key: target_file_2
        }

        config_file_changes: ConfigFileChanges = determine_config_file_changes(map_of_local_config_files, map_of_target_config_files)

        self.assertFalse(config_file_changes.list_of_files_that_are_deprecated)
        self.assertFalse(config_file_changes.list_of_files_to_rename)
        self.assertCountEqual(['test.dat', 'another.txt'], config_file_changes.list_of_files_to_leave)
        self.assertFalse(config_file_changes.list_of_new_files)

    def test_two_local_files_and_two_target_files_where_one_target_file_has_same_name_but_different_key(self) -> None:
        """
        ### Local
        * 23: test.dat
        * 8: another.txt
        ### Target
        * 23: test.dat
        * 31: another.txt
        ### Expected
        * files_to_leave  = test.dat
        * files_to_remove = another.txt
        * new_files       = another.txt
        """
        local_file_1 = StrictMock(IConfigFile)
        local_file_1.config_key = PropertyMock(return_value=23)
        local_file_1.filename = PropertyMock(return_value='test.dat')
        local_file_2 = StrictMock(IConfigFile)
        local_file_2.config_key = PropertyMock(return_value=8)
        local_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_local_config_files: dict[int, IConfigFile] = {
            local_file_1.config_key: local_file_1,
            local_file_2.config_key: local_file_2
        }
        target_file_1 = StrictMock(IConfigFile)
        target_file_1.config_key = PropertyMock(return_value=23)
        target_file_1.filename = PropertyMock(return_value='test.dat')
        target_file_2 = StrictMock(IConfigFile)
        target_file_2.config_key = PropertyMock(return_value=31)
        target_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_target_config_files: dict[int, IConfigFile] = {
            target_file_1.config_key: target_file_1,
            target_file_2.config_key: target_file_2
        }

        config_file_changes: ConfigFileChanges = determine_config_file_changes(map_of_local_config_files, map_of_target_config_files)

        self.assertCountEqual(['another.txt'], config_file_changes.list_of_files_that_are_deprecated)
        self.assertFalse(config_file_changes.list_of_files_to_rename)
        self.assertCountEqual(['test.dat'], config_file_changes.list_of_files_to_leave)
        self.assertCountEqual(['another.txt'], config_file_changes.list_of_new_files)

    def test_two_local_files_and_two_target_files_where_one_target_file_has_same_key_but_different_name(self) -> None:
        """
        ### Local
        * 23: test.dat
        * 8: another.txt
        ### Target
        * 23: test.dat
        * 8: something.log
        ### Expected
        * files_to_leave  = test.dat
        * files_to_rename = another.txt -> something.log
        """
        local_file_1 = StrictMock(IConfigFile)
        local_file_1.config_key = PropertyMock(return_value=23)
        local_file_1.filename = PropertyMock(return_value='test.dat')
        local_file_2 = StrictMock(IConfigFile)
        local_file_2.config_key = PropertyMock(return_value=8)
        local_file_2.filename = PropertyMock(return_value='another.txt')
        map_of_local_config_files: dict[int, IConfigFile] = {
            local_file_1.config_key: local_file_1,
            local_file_2.config_key: local_file_2
        }
        target_file_1 = StrictMock(IConfigFile)
        target_file_1.config_key = PropertyMock(return_value=23)
        target_file_1.filename = PropertyMock(return_value='test.dat')
        target_file_2 = StrictMock(IConfigFile)
        target_file_2.config_key = PropertyMock(return_value=8)
        target_file_2.filename = PropertyMock(return_value='something.log')
        map_of_target_config_files: dict[int, IConfigFile] = {
            target_file_1.config_key: target_file_1,
            target_file_2.config_key: target_file_2
        }

        config_file_changes: ConfigFileChanges = determine_config_file_changes(map_of_local_config_files, map_of_target_config_files)

        self.assertFalse(config_file_changes.list_of_files_that_are_deprecated)
        self.assertEqual(1, len(config_file_changes.list_of_files_to_rename))
        self.assertEqual('another.txt', config_file_changes.list_of_files_to_rename[0].original_filename)
        self.assertEqual('something.log', config_file_changes.list_of_files_to_rename[0].new_filename)
        self.assertCountEqual(['test.dat'], config_file_changes.list_of_files_to_leave)
        self.assertFalse(config_file_changes.list_of_new_files)

class TestDetermineNewerVersions(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.variant_windows = StrictMock(IVariant)
        self.variant_linux = StrictMock(IVariant)
        self.variant_vms = StrictMock(IVariant)
        self.variant_windows.name = PropertyMock(return_value='windows')
        self.variant_linux.name = PropertyMock(return_value='linux')
        self.variant_vms.name = PropertyMock(return_value='vms')

        self.version_number_0 = VersionNumber.parse('1.2.3')
        self.version_number_1 = VersionNumber.parse('2.0.1')
        self.version_number_2 = VersionNumber.parse('3.7.9')
        self.version_number_3 = VersionNumber.parse('4.1.2')
        self.version_number_4 = VersionNumber.parse('5.6.8')

        self.version_1 = StrictMock(IVersion)
        self.version_2 = StrictMock(IVersion)
        self.version_3 = StrictMock(IVersion)
        self.version_1.number = PropertyMock(return_value=self.version_number_1)
        self.version_2.number = PropertyMock(return_value=self.version_number_2)
        self.version_3.number = PropertyMock(return_value=self.version_number_3)

        self.variant_windows.versions = PropertyMock(return_value={})
        self.variant_linux.versions = PropertyMock(return_value={
            self.version_3.number: self.version_3,
            self.version_1.number: self.version_1,
            self.version_2.number: self.version_2
        })
        self.variant_vms.versions = PropertyMock(return_value={})

    def test_when_starting_variant_does_not_match_target_variant_returns_no_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', StrictMock(IVersionNumber), self.variant_windows, StrictMock(IVersionNumber))
        self.assertFalse(list_of_versions)

    def test_when_matching_variant_has_no_versions_returns_no_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('vms', StrictMock(IVersionNumber), self.variant_vms, StrictMock(IVersionNumber))
        self.assertFalse(list_of_versions)

    def test_starting_version_is_first_and_target_version_is_last_returns_all_later_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_1, self.variant_linux, self.version_number_3)
        self.assertListEqual([self.version_2, self.version_3], list_of_versions)

    def test_starting_version_is_second_and_target_version_is_last_returns_all_later_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_2, self.variant_linux, self.version_number_3)
        self.assertListEqual([self.version_3], list_of_versions)

    def test_starting_version_is_last_and_same_as_target_version_returns_no_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_3, self.variant_linux, self.version_number_3)
        self.assertListEqual([], list_of_versions)

    def test_starting_version_is_first_and_target_version_is_second_returns_target_version(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_1, self.variant_linux, self.version_number_2)
        self.assertListEqual([self.version_2], list_of_versions)

    def test_starting_version_is_first_and_same_as_target_version_returns_no_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_1, self.variant_linux, self.version_number_1)
        self.assertListEqual([], list_of_versions)

    def test_starting_version_is_not_list_and_lowest_and_target_version_is_last_returns_all_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_0, self.variant_linux, self.version_number_3)
        self.assertListEqual([self.version_1, self.version_2, self.version_3], list_of_versions)

    def test_starting_version_is_first_and_target_version_is_not_in_list_and_highest_returns_all_later_versions(self) -> None:
        list_of_versions: list[IVersion] = determine_newer_versions('linux', self.version_number_1, self.variant_linux, self.version_number_4)
        self.assertListEqual([self.version_2, self.version_3], list_of_versions)

class TestDetermineFilesToBackup(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.core_file_1 = StrictMock(ICoreFile)
        self.core_file_1.filename = PropertyMock(return_value='file1')
        self.core_file_2 = StrictMock(ICoreFile)
        self.core_file_2.filename = PropertyMock(return_value='file2')

        self.config_filename_1 = 'config_file1'
        self.config_filename_2 = 'config_file2'
        self.config_new_filename_1 = 'config_new_file1'
        self.config_new_filename_2 = 'config_new_file2'

    def test_no_core_files_or_config_fields_yields_empty_list(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_new_core_files_yields_empty_list(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [self.core_file_1, self.core_file_2],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )
        file_tester = StrictMock(IFileTester)
        file_tester.stat_file = Mock(side_effect=lambda filename: {
            self.core_file_1.filename: FileStat(self.core_file_1.filename, False, False),
            self.core_file_2.filename: FileStat(self.core_file_1.filename, False, False)
        }[filename])

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, file_tester)

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_new_core_files_with_destination_file_clash_yields_clobbered_file_list(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [self.core_file_1, self.core_file_2],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )
        file_tester = StrictMock(IFileTester)
        file_tester.stat_file = Mock(side_effect=lambda filename: {
            self.core_file_1.filename: FileStat(self.core_file_1.filename, False, False),
            self.core_file_2.filename: FileStat(self.core_file_2.filename, True, True)
        }[filename])

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, file_tester)

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertCountEqual([self.core_file_2.filename], backup_file_set.list_of_clobbered_files_to_backup)

    def test_new_core_files_with_desitnation_directory_clash_raises_error(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [self.core_file_1, self.core_file_2],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )
        file_tester = StrictMock(IFileTester)
        file_tester.stat_file = Mock(side_effect=lambda filename: {
            self.core_file_1.filename: FileStat(self.core_file_1.filename, False, False),
            self.core_file_2.filename: FileStat(self.core_file_1.filename, True, False)
        }[filename])

        with self.assertRaises(UpdateStrategyError) as ex:
            determine_files_to_backup(core_file_changes, config_file_changes, file_tester)

        self.assertEqual(f"Destination file '{self.core_file_2.filename}' already exists and is a directory", str(ex.exception))

    def test_missing_core_files_yields_empty_list(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [self.core_file_1, self.core_file_2],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_changed_core_files_yields_list_of_files(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [self.core_file_1, self.core_file_2],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertCountEqual([self.core_file_1.filename, self.core_file_2.filename], backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_unchanged_core_files_yields_list_of_files(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [self.core_file_1, self.core_file_2],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertCountEqual([self.core_file_1.filename, self.core_file_2.filename], backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_removed_core_files_yields_list_of_files(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = [self.core_file_1.filename, self.core_file_2.filename]
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertCountEqual([self.core_file_1.filename, self.core_file_2.filename], backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_new_config_files_yields_empty_list(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = [self.config_filename_1, self.config_filename_2]
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertFalse(backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_leave_config_files_yields_list_of_files(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [self.config_filename_1, self.config_filename_2],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertCountEqual([self.config_filename_1, self.config_filename_2], backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_rename_config_files_yields_list_of_files(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [],
            list_of_files_to_rename = [FileRename(self.config_filename_1, self.config_new_filename_1), FileRename(self.config_filename_2, self.config_new_filename_2)],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertCountEqual([self.config_filename_1, self.config_filename_2], backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

    def test_remove_config_files_yields_list_of_files(self) -> None:
        core_file_changes: CoreFileChanges = CoreFileChanges(
            list_of_new_files_to_acquire = [],
            list_of_missing_files_to_acquire = [],
            list_of_changed_files_to_acquire = [],
            list_of_unchanged_files = [],
            list_of_files_that_are_deprecated = []
        )
        config_file_changes: ConfigFileChanges = ConfigFileChanges(
            list_of_files_that_are_deprecated = [self.config_filename_1, self.config_filename_2],
            list_of_files_to_rename = [],
            list_of_files_to_leave = [],
            list_of_new_files = []
        )

        backup_file_set: BackupFileSet = determine_files_to_backup(core_file_changes, config_file_changes, StrictMock(IFileTester))

        self.assertFalse(backup_file_set.list_of_core_files_to_backup)
        self.assertCountEqual([self.config_filename_1, self.config_filename_2], backup_file_set.list_of_config_files_to_backup)
        self.assertFalse(backup_file_set.list_of_clobbered_files_to_backup)

class TestDetermineUpdateStrategy(unittest.TestCase):
    """
    The following configuration is used to test the integrated update strategy.

    Local Manifest

        Version: windows / 1.0.0

    Target

    Core files

        Local Manifest  |  Current Files / Digest  |  Target Files / Digest  |  State      / Action
        ----------------|--------------------------|-------------------------|------------------------------
        apple.txt       |  apple.txt     / A       |  apple.txt    / A       |  Unchanged  / Backup
        banana.dat      |  banana.dat    / B       |  banana.dat   / F       |  Changed    / Backup, Acquire
        pear.log        |  pear.log      / C       |                         |  Deprecated / Backup, Delete
                        |                          |  cheese.net   / G       |  New        / Acquire
        mainApp         |                          |  mainApp      / D       |  Missing    / Acquire
        mainApp.exe     |  mainApp.exe   / E       |  mainApp.exe  / H       |  Changed    / Backup, Acquire
                        |  database.id   / I       |  database.id  / J       |  Clobbered  / Backup, Acquire

    Config files

        Local Manifest / Key  |  Current Files  |  Target Files / Key  |  State      / Action
        ----------------------|-----------------|----------------------|-----------------------------
        config.json    / 3    |  config.json    |  config.json  / 3    |  Unchanged  / Backup
        history.txt    / 5    |  history.txt    |  history.log  / 5    |  Renamed    / Backup, Rename
        auto.txt       / 4    |  auto.txt       |                      |  Deprecated / Backup, Delete
                              |                 |  logging.txt  / 6    |  New        / -

    Launch files

        Operating System  |  Target File
        ------------------|-------------
        win32             |  mainApp.exe
        linux             |  mainApp
    """
    def setUp(self) -> None: # pylint: disable=too-many-statements
        super().setUp()

        self.local_version_number = VersionNumber.parse('1.0.0')
        self.target_version_number = VersionNumber.parse('2.0.0')

        self.local_manifest = StrictMock(ILocalManifest,
            variant_name = PropertyMock(return_value='windows'),
            version_number = PropertyMock(return_value=self.local_version_number),
            core_files = PropertyMock(return_value=[
                'apple.txt',
                'banana.dat',
                'pear.log',
                'mainApp',
                'mainApp.exe'
            ]),
            config_files = PropertyMock(return_value={
                3: StrictMock(IConfigFile,
                    config_key = PropertyMock(return_value=3),
                    filename = PropertyMock(return_value='config.json')
                ),
                5: StrictMock(IConfigFile,
                    config_key = PropertyMock(return_value=5),
                    filename = PropertyMock(return_value='history.txt')
                ),
                4: StrictMock(IConfigFile,
                    config_key = PropertyMock(return_value=4),
                    filename = PropertyMock(return_value='auto.txt')
                )
            })
        )

        self.version_1 = StrictMock(IVersion,
            number = PropertyMock(return_value=self.local_version_number)
        )

        self.target_core_file_1 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='apple.txt'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_A)
        )
        self.target_core_file_2 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='banana.dat'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_F)
        )
        self.target_core_file_3 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='cheese.net'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_G)
        )
        self.target_core_file_4 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='mainApp'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_D)
        )
        self.target_core_file_5 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='mainApp.exe'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_H)
        )
        self.target_core_file_6 = StrictMock(ICoreFile,
            filename = PropertyMock(return_value='database.id'),
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_I)
        )

        self.target_config_file_1 = StrictMock(IConfigFile,
            config_key = PropertyMock(return_value=3),
            filename = PropertyMock(return_value='config.json')
        )
        self.target_config_file_2 = StrictMock(IConfigFile,
            config_key = PropertyMock(return_value=5),
            filename = PropertyMock(return_value='history.log')
        )
        self.target_config_file_3 = StrictMock(IConfigFile,
            config_key = PropertyMock(return_value=6),
            filename = PropertyMock(return_value='logging.txt')
        )

        self.launch_file_1 = StrictMock(ILaunchFile,
            operating_system = PropertyMock(return_value='win32'),
            filename = PropertyMock(return_value='mainApp.exe')
        )
        self.launch_file_2 = StrictMock(ILaunchFile,
            operating_system = PropertyMock(return_value='linux'),
            filename = PropertyMock(return_value='mainApp')
        )

        self.version_2 = StrictMock(IVersion,
            number = PropertyMock(return_value=self.target_version_number),
            core_files = PropertyMock(return_value={
                self.target_core_file_1.filename: self.target_core_file_1,
                self.target_core_file_2.filename: self.target_core_file_2,
                self.target_core_file_3.filename: self.target_core_file_3,
                self.target_core_file_4.filename: self.target_core_file_4,
                self.target_core_file_5.filename: self.target_core_file_5,
                self.target_core_file_6.filename: self.target_core_file_6
            }),
            config_files = PropertyMock(return_value={
                self.target_config_file_1.config_key: self.target_config_file_1,
                self.target_config_file_2.config_key: self.target_config_file_2,
                self.target_config_file_3.config_key: self.target_config_file_3
            }),
            launch_files = PropertyMock(return_value={
                self.launch_file_1.operating_system: self.launch_file_1,
                self.launch_file_2.operating_system: self.launch_file_2
            })
        )

        self.variant_main = StrictMock(IVariant,
            name = PropertyMock(return_value='main'),
            versions = PropertyMock(return_value={
                self.version_1.number: self.version_1,
                self.version_2.number: self.version_2
            })
        )

        self.version_manifest = StrictMock(IVersionManifest,
            variants = PropertyMock(return_value={
                self.variant_main.name: self.variant_main
            })
        )

        asset_file_a = StrictMock(IAsset,
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_A)
        )
        asset_file_d = StrictMock(IAsset,
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_D)
        )
        asset_file_f = StrictMock(IAsset,
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_F)
        )
        asset_file_g = StrictMock(IAsset,
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_G)
        )
        asset_file_h = StrictMock(IAsset,
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_H)
        )
        asset_file_i = StrictMock(IAsset,
            digest = PropertyMock(return_value=FAKE_HEXDIGEST_I)
        )
        self.asset_register = StrictMock(IAssetRegister,
            assets = PropertyMock(return_value={
                asset_file_a.digest: asset_file_a,
                asset_file_d.digest: asset_file_d,
                asset_file_f.digest: asset_file_f,
                asset_file_g.digest: asset_file_g,
                asset_file_h.digest: asset_file_h,
                asset_file_i.digest: asset_file_i
            })
        )

        def mock_hash_file(filename: str) -> bytes:
            digest: bytes | None = {
                'apple.txt': FAKE_HEXDIGEST_A.digest_bytes,
                'banana.dat': FAKE_HEXDIGEST_B.digest_bytes,
                'pear.log': FAKE_HEXDIGEST_C.digest_bytes,
                'mainApp.exe': FAKE_HEXDIGEST_E.digest_bytes
            }.get(filename)
            if not digest:
                raise HashFileNotFoundError()
            return digest

        self.file_hasher = StrictMock(IFileHasher,
            hash_file = Mock(side_effect=mock_hash_file)
        )

        def mock_stat_file(filename: str) -> FileStat:
            file_stat: FileStat | None = {
                'apple.txt': FileStat('apple.txt', True, True),
                'banana.dat': FileStat('banana.dat', True, True),
                'pear.log': FileStat('pear.log', True, True),
                'mainApp.exe': FileStat('mainApp.exe', True, True),
                'database.id': FileStat('database.id', True, True),
                'config.json': FileStat('config.json', True, True),
                'history.txt': FileStat('history.txt', True, True),
                'auto.txt': FileStat('auto.txt', True, True)
            }.get(filename)
            if not file_stat:
                return FileStat(filename, False, False)
            return file_stat

        self.file_tester = StrictMock(IFileTester,
            stat_file = Mock(side_effect=mock_stat_file)
        )

    @patch('sys.platform', 'win32')
    @patch('sys.executable', r'C:\Temp\milodb\mainApp.exe')
    def test_integration_of_all_functions_on_windows(self) -> None:
        mock_path.map_of_absolute_resolution = {
            Path(r'C:\Temp\milodb'): Path(r'C:\Temp\milodb'),
            Path(r'C:\Temp\milodb\mainApp.exe'): Path(r'C:\Temp\milodb\mainApp.exe')
        }
        mock_path.map_of_relative_to_resolution = {
            (Path(r'C:\Temp\milodb\mainApp.exe'), Path(r'C:\Temp\milodb')): Path('mainApp.exe')
        }

        with patch.object(Path, 'absolute', mock_path.absolute):
            with patch.object(Path, 'relative_to', mock_path.relative_to):
                self._test_integration_of_all_functions(r'C:\Temp\milodb', 'win32', 'mainApp.exe')

    @patch('sys.platform', 'linux')
    @patch('sys.executable', r'/tmp/milodb/mainApp')
    def test_integration_of_all_functions_on_linux(self) -> None:
        mock_path.map_of_absolute_resolution = {
            Path(r'/tmp/milodb'): Path(r'/tmp/milodb'),
            Path(r'/tmp/milodb/mainApp'): Path(r'/tmp/milodb/mainApp')
        }
        mock_path.map_of_relative_to_resolution = {
            (Path(r'/tmp/milodb/mainApp'), Path(r'/tmp/milodb')): Path('mainApp')
        }

        with patch.object(Path, 'absolute', mock_path.absolute):
            with patch.object(Path, 'relative_to', mock_path.relative_to):
                self._test_integration_of_all_functions('/tmp/milodb', 'linux', 'mainApp')

    @patch('sys.platform', 'osx')
    @patch('sys.executable', r'/tmp/milodb/mainApp')
    def test_integration_of_all_functions_on_osx(self) -> None:
        mock_path.map_of_absolute_resolution = {
            Path(r'/tmp/milodb'): Path(r'/tmp/milodb'),
            Path(r'/tmp/milodb/mainApp'): Path(r'/tmp/milodb/mainApp')
        }
        mock_path.map_of_relative_to_resolution = {
            (Path(r'/tmp/milodb/mainApp'), Path(r'/tmp/milodb')): Path('mainApp')
        }

        with patch.object(Path, 'absolute', mock_path.absolute):
            with patch.object(Path, 'relative_to', mock_path.relative_to):
                self._test_integration_of_all_functions('/tmp/milodb', 'osx', None)

    def _test_integration_of_all_functions(self, root_directory: str, os_name: str, app_name: str | None) -> None:
        requested_version_number: IVersionNumber | None = None
        requested_variant: str | None = 'main'

        update_strategy: UpdateStrategy = determine_update_strategy(
            root_directory, requested_version_number, requested_variant,
            self.local_manifest, self.version_manifest, self.asset_register, self.file_hasher, self.file_tester)

        self.assertEqual(root_directory, update_strategy.root_directory)
        self.assertIs(self.local_manifest, update_strategy.local_manifest)
        self.assertIs(self.version_manifest, update_strategy.version_manifest)

        self.assertCountEqual([self.target_core_file_3, self.target_core_file_6], update_strategy.core_file_changes.list_of_new_files_to_acquire)
        self.assertCountEqual([self.target_core_file_4], update_strategy.core_file_changes.list_of_missing_files_to_acquire)
        self.assertCountEqual([self.target_core_file_2, self.target_core_file_5], update_strategy.core_file_changes.list_of_changed_files_to_acquire)
        self.assertCountEqual([self.target_core_file_1], update_strategy.core_file_changes.list_of_unchanged_files)
        self.assertCountEqual(['pear.log'], update_strategy.core_file_changes.list_of_files_that_are_deprecated)

        self.assertCountEqual(['auto.txt'], update_strategy.config_file_changes.list_of_files_that_are_deprecated)
        self.assertCountEqual([FileRename('history.txt', 'history.log')], update_strategy.config_file_changes.list_of_files_to_rename)
        self.assertCountEqual(['config.json'], update_strategy.config_file_changes.list_of_files_to_leave)
        self.assertCountEqual(['logging.txt'], update_strategy.config_file_changes.list_of_new_files)

        self.assertCountEqual(['apple.txt', 'banana.dat', 'pear.log', 'mainApp.exe'], update_strategy.backup_file_set.list_of_core_files_to_backup)
        self.assertCountEqual(['config.json', 'history.txt', 'auto.txt'], update_strategy.backup_file_set.list_of_config_files_to_backup)
        self.assertCountEqual(['database.id'], update_strategy.backup_file_set.list_of_clobbered_files_to_backup)

        self.assertEqual(os_name, update_strategy.current_operating_system)
        if os_name and app_name:
            self.assertIsNotNone(update_strategy.launch_executable)
            if update_strategy.launch_executable:
                self.assertEqual(os_name, update_strategy.launch_executable.operating_system)
                self.assertEqual(app_name, update_strategy.launch_executable.filename)
        else:
            self.assertIsNone(update_strategy.launch_executable)
