# MiloDB by FrozenWolf is marked with CC0 1.0.
# To view a copy of this license, visit http://creativecommons.org/publicdomain/zero/1.0
import dataclasses
from abc import ABC
from types import UnionType
from typing import Any, get_origin
from unittest.mock import Mock, NonCallableMock, PropertyMock

class StrictMock(NonCallableMock):
    def __init__(self, interface: type[ABC], **kwargs: Any) -> None:
        self._mocked_interface: type[ABC] = interface
        super().__init__(spec=interface, name=interface.__name__)
        arg_name: str
        arg_value: Any
        for arg_name, arg_value in kwargs.items():
            self.__setattr__(arg_name, arg_value)

    def __getattr__(self, name: str) -> Any:
        if name in dir():
            return super().__getattr__(name)
        if name in dir(self._mocked_interface):
            raise AssertionError(f"Mock interface '{self._mocked_interface.__name__}' unexpected access to '{name}'")
        raise AttributeError(f"Mock interface '{self._mocked_interface.__name__}' has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        if name == '_mocked_interface':
            object.__setattr__(self, name, value)
        elif name in dir(self._mocked_interface):
            if isinstance(getattr(self._mocked_interface, name), property):
                assert isinstance(value, PropertyMock), f"Member '{self._mocked_interface.__name__}.{name}' is a property and should be assigned to a PropertyMock"
                setattr(type(self), name, value)
            elif callable(getattr(self._mocked_interface, name)):
                assert (isinstance(value, Mock) or callable(value)) and not isinstance(value, PropertyMock), f"Member '{self._mocked_interface.__name__}.{name}' is a method and should be assigned to a Mock"
                super().__setattr__(name, value)
            else:
                super().__setattr__(name, value)
        else:
            raise AttributeError(f"Mock interface '{self._mocked_interface.__name__}' has no attribute '{name}'")

class StrictDataclassMock(NonCallableMock):
    def __init__(self, cls: type[Any], **kwargs: Any) -> None:
        if not dataclasses.is_dataclass(cls):
            raise AssertionError(f"Class '{cls.__name__}' is not a dataclass")
        self._mocked_class: type[Any] = cls
        super().__init__(spec=cls, name=cls.__name__)
        arg_name: str
        arg_value: Any
        for arg_name, arg_value in kwargs.items():
            self.__setattr__(arg_name, arg_value)

    def __getattr__(self, name: str) -> Any:
        if name in dir():
            return super().__getattr__(name)
        if name in self._mocked_class.__annotations__:
            raise AssertionError(f"Mock dataclass '{self._mocked_class.__name__}' unexpected access to '{name}'")
        raise AttributeError(f"Mock dataclass '{self._mocked_class.__name__}' has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        if name == '_mocked_class':
            object.__setattr__(self, name, value)
        else:
            member_type: type[Any] | None = self._mocked_class.__annotations__.get(name)
            if not member_type:
                raise AttributeError(f"Mock interface '{self._mocked_class.__name__}' has no attribute '{name}'")
            base_type: type[Any] | None = get_origin(member_type)
            if base_type is UnionType:
                base_type = None
            if not isinstance(value, base_type or member_type):
                raise AttributeError(f"Mock interface '{self._mocked_class.__name__}' attribute '{name}' is being assigned the wrong type")
            super().__setattr__(name, value)
